@@ -12,6 +12,7 @@ use File::Compare ();
 use Data::Dumper ();
 use IO::File ();
 
+#################### Constructors ###########################
 sub new {
   my $package = shift;
   my %input = @_;
@@ -76,6 +77,37 @@ sub new {
   return $self;
 }
 
+sub resume {
+  my $package = shift;
+  my $self = bless {@_}, $package;
+  
+  $self->read_config;
+  
+  my $perl = $self->find_perl_interpreter;
+  warn(" * WARNING: Configuration was initially created with '$self->{properties}{perl}',\n".
+       "   but we are now using '$perl'.\n")
+    unless $perl eq $self->{properties}{perl};
+  
+  $self->cull_args(@ARGV);
+  $self->{action} ||= 'build';
+  
+  return $self;
+}
+
+sub instance {
+  my $package = shift;
+  # hmm, wonder what the right thing to do here is
+  local @ARGV;
+  return $package->resume(
+			  properties => {
+					 config_dir => '_build',
+					 build_script => 'Build',
+					},
+			 );
+}
+
+################## End constructors #########################
+
 sub _set_install_paths {
   my $self = shift;
   my $c = $self->{config};
@@ -168,21 +200,15 @@ sub y_n {
   }
 }
 
-sub resume {
-  my $package = shift;
-  my $self = bless {@_}, $package;
+sub notes {
+  my $self = shift;
+  return $self->_persistent_hash_read('notes') unless @_;
   
-  $self->read_config;
+  my $key = shift;
+  return $self->_persistent_hash_read('notes', $key) unless @_;
   
-  my $perl = $self->find_perl_interpreter;
-  warn(" * WARNING: Configuration was initially created with '$self->{properties}{perl}',\n".
-       "   but we are now using '$perl'.\n")
-    unless $perl eq $self->{properties}{perl};
-  
-  $self->cull_args(@ARGV);
-  $self->{action} ||= 'build';
-  
-  return $self;
+  my $value = shift;
+  return $self->_persistent_hash_write('notes', { $key => $value });
 }
 
 {
@@ -388,44 +414,74 @@ sub version_from_file {
   return $result;
 }
 
-sub _write_cleanup {
-  my $self = shift;
-  my @to_write = keys %{ $self->{add_to_cleanup} }
-    or return;
+sub _persistent_hash_write {
+  my ($self, $name, $href) = @_;
+  $href ||= {};
+  my $ph = $self->{phash}{$name} ||= {disk => {}, new => {}};
   
-  my $file = $self->config_file('cleanup');
-  my $fh = IO::File->new(">> $file") or die "Can't write to $file: $!";
-  print $fh "$_\n" foreach @to_write;
-  close $fh;
+  @{$ph->{new}}{ keys %$href } = values %$href;  # Merge
+
+  # Do some optimization to avoid unnecessary writes
+  foreach my $key (keys %{ $ph->{new} }) {
+    next if ref $ph->{new}{$key};
+    next if ref $ph->{disk}{$key} or !exists $ph->{disk}{$key};
+    delete $ph->{new}{$key} if $ph->{new}{$key} eq $ph->{disk}{$key};
+  }
   
-  @{ $self->{cleanup} }{ @to_write } = ();
-  $self->{add_to_cleanup} = {};
+  if (my $file = $self->config_file($name)) {
+    return if -e $file and !keys %{ $ph->{new} };  # Nothing to do
+    
+    local $Data::Dumper::Terse = 1;
+    my $fh = IO::File->new("> $file") or die "Can't write to $file: $!";
+    @{$ph->{disk}}{ keys %{$ph->{new}} } = values %{$ph->{new}};  # Merge
+    print $fh Data::Dumper::Dumper($ph->{disk});
+    close $fh;
+    
+    $ph->{new} = {};
+  }
+  return $self->_persistent_hash_read($name);
 }
 
-sub cleanup_is_flushed {
+sub _persistent_hash_read {
   my $self = shift;
-  return ! keys %{ $self->{add_to_cleanup} };
+  my $name = shift;
+  my $ph = $self->{phash}{$name} ||= {disk => {}, new => {}};
+
+  if (@_) {
+    # Return 1 key as a scalar
+    my $key = shift;
+    return $ph->{new}{$key} if exists $ph->{new}{$key};
+    return $ph->{disk}{$key};
+  } else {
+    # Return all data
+    my $out = (keys %{$ph->{new}}
+	       ? {%{$ph->{disk}}, %{$ph->{new}}}
+	       : $ph->{disk});
+    return wantarray ? %$out : $out;
+  }
+}
+
+sub _persistent_hash_restore {
+  my ($self, $name) = @_;
+  my $ph = $self->{phash}{$name} ||= {disk => {}, new => {}};
+  
+  my $file = $self->config_file($name) or die "No config file '$name'";
+  my $fh = IO::File->new("< $file") or die "Can't read $file: $!";
+  
+  $ph->{disk} = eval do {local $/; <$fh>};
+  die $@ if $@;
 }
 
 sub add_to_cleanup {
   my $self = shift;
-
-  # $self->{cleanup} contains files that are already written in the
-  # 'cleanup' file.  $self->{add_to_cleanup} is a buffer that we
-  # haven't written yet (and may never write if we don't ever create
-  # the cleanup file).
-  
-  my @new_files = grep {!exists $self->{cleanup}{$_}} @_
-    or return;
-  
-  @{$self->{add_to_cleanup}}{ @new_files } = ();
-  
-  $self->_write_cleanup if $self->config_file('cleanup');
+  my %files = map {$_, 1} @_;
+  $self->_persistent_hash_write('cleanup', \%files);
 }
 
 sub cleanup {
   my $self = shift;
-  return (keys %{$self->{cleanup}}, keys %{$self->{add_to_cleanup}});
+  my $all = $self->_persistent_hash_read('cleanup');
+  return keys %$all;
 }
 
 sub config_file {
@@ -443,14 +499,10 @@ sub read_config {
   die if $@;
   ($self->{args}, $self->{config}, $self->{properties}) = @$ref;
   close $fh;
-  
-  my $cleanup_file = $self->config_file('cleanup');
-  $self->{cleanup} = {};
-  if (-e $cleanup_file) {
-    my $fh = IO::File->new($cleanup_file) or die "Can't read '$file': $!";
-    my @files = <$fh>;
-    chomp @files;
-    @{$self->{cleanup}}{@files} = ();
+
+  for ('cleanup', 'notes') {
+    next unless -e $self->config_file($_);
+    $self->_persistent_hash_restore($_);
   }
 }
 
@@ -472,8 +524,8 @@ sub write_config {
   my @items = qw(requires build_requires conflicts recommends);
   print $fh Data::Dumper::Dumper( { map { $_, $self->$_() } @items } );
   close $fh;
-  
-  $self->_write_cleanup;
+
+  $self->_persistent_hash_write('cleanup');
 }
 
 sub requires       { shift()->{properties}{requires} }
@@ -1101,7 +1153,6 @@ sub ACTION_builddocs {
   require Pod::Man;
   $self->manify_bin_pods();
   $self->manify_lib_pods();
-  return $self;
 }
 
 sub manify_bin_pods {
@@ -1120,8 +1171,6 @@ sub manify_bin_pods {
     $parser->parse_from_file( $file, $outfile );
     $files->{$file} = $outfile;
   }
-
-  return $self;
 }
 
 sub manify_lib_pods {
@@ -1140,8 +1189,6 @@ sub manify_lib_pods {
     $parser->parse_from_file( $file, $outfile );
     $files->{$file} = $outfile;
   }
-
-  return $self;
 }
 
 sub _find_pods {
@@ -1158,14 +1205,14 @@ sub _find_pods {
 
 sub contains_pod {
   my ($self, $file) = @_;
-  return 0 unless -T $file;  # Only look at text files
+  return '' unless -T $file;  # Only look at text files
   
   my $fh = IO::File->new( $file ) or die "Can't open $file: $!";
   while (my $line = <$fh>) {
     return 1 if $line =~ /^\=(?:head|pod|item)/;
   }
   
-  return 0;
+  return '';
 }
 
 # Adapted from ExtUtils::MM_Unix
